@@ -16,16 +16,11 @@
 # You should have received a copy of the GNU Lesser General Public License
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
-from __future__ import print_function
-
 import os
 import pickle
 import tempfile
-
-try:
-    import unittest
-except ImportError:
-    import unittest2 as unittest
+import json
+import unittest
 
 from httmock import HTTMock  # noqa
 from httmock import response  # noqa
@@ -403,17 +398,31 @@ class TestGitlabAuth(unittest.TestCase):
         gl = Gitlab("http://localhost", private_token="private_token", api_version="4")
         self.assertEqual(gl.private_token, "private_token")
         self.assertEqual(gl.oauth_token, None)
+        self.assertEqual(gl.job_token, None)
         self.assertEqual(gl._http_auth, None)
-        self.assertEqual(gl.headers["PRIVATE-TOKEN"], "private_token")
         self.assertNotIn("Authorization", gl.headers)
+        self.assertEqual(gl.headers["PRIVATE-TOKEN"], "private_token")
+        self.assertNotIn("JOB-TOKEN", gl.headers)
 
     def test_oauth_token_auth(self):
         gl = Gitlab("http://localhost", oauth_token="oauth_token", api_version="4")
         self.assertEqual(gl.private_token, None)
         self.assertEqual(gl.oauth_token, "oauth_token")
+        self.assertEqual(gl.job_token, None)
         self.assertEqual(gl._http_auth, None)
         self.assertEqual(gl.headers["Authorization"], "Bearer oauth_token")
         self.assertNotIn("PRIVATE-TOKEN", gl.headers)
+        self.assertNotIn("JOB-TOKEN", gl.headers)
+
+    def test_job_token_auth(self):
+        gl = Gitlab("http://localhost", job_token="CI_JOB_TOKEN", api_version="4")
+        self.assertEqual(gl.private_token, None)
+        self.assertEqual(gl.oauth_token, None)
+        self.assertEqual(gl.job_token, "CI_JOB_TOKEN")
+        self.assertEqual(gl._http_auth, None)
+        self.assertNotIn("Authorization", gl.headers)
+        self.assertNotIn("PRIVATE-TOKEN", gl.headers)
+        self.assertEqual(gl.headers["JOB-TOKEN"], "CI_JOB_TOKEN")
 
     def test_http_auth(self):
         gl = Gitlab(
@@ -425,6 +434,7 @@ class TestGitlabAuth(unittest.TestCase):
         )
         self.assertEqual(gl.private_token, "private_token")
         self.assertEqual(gl.oauth_token, None)
+        self.assertEqual(gl.job_token, None)
         self.assertIsInstance(gl._http_auth, requests.auth.HTTPBasicAuth)
         self.assertEqual(gl.headers["PRIVATE-TOKEN"], "private_token")
         self.assertNotIn("Authorization", gl.headers)
@@ -435,8 +445,6 @@ class TestGitlab(unittest.TestCase):
         self.gl = Gitlab(
             "http://localhost",
             private_token="private_token",
-            email="testuser@test.com",
-            password="testpassword",
             ssl_verify=True,
             api_version=4,
         )
@@ -449,66 +457,7 @@ class TestGitlab(unittest.TestCase):
         self.assertTrue(hasattr(unpickled, "_objects"))
         self.assertEqual(unpickled._objects, original_gl_objects)
 
-    def test_credentials_auth_nopassword(self):
-        self.gl.email = None
-        self.gl.password = None
-
-        @urlmatch(
-            scheme="http", netloc="localhost", path="/api/v4/session", method="post"
-        )
-        def resp_cont(url, request):
-            headers = {"content-type": "application/json"}
-            content = '{"message": "message"}'.encode("utf-8")
-            return response(404, content, headers, None, 5, request)
-
-        with HTTMock(resp_cont):
-            self.assertRaises(GitlabHttpError, self.gl._credentials_auth)
-
-    def test_credentials_auth_notok(self):
-        @urlmatch(
-            scheme="http", netloc="localhost", path="/api/v4/session", method="post"
-        )
-        def resp_cont(url, request):
-            headers = {"content-type": "application/json"}
-            content = '{"message": "message"}'.encode("utf-8")
-            return response(404, content, headers, None, 5, request)
-
-        with HTTMock(resp_cont):
-            self.assertRaises(GitlabHttpError, self.gl._credentials_auth)
-
-    def test_auth_with_credentials(self):
-        self.gl.private_token = None
-        self.test_credentials_auth(callback=self.gl.auth)
-
-    def test_auth_with_token(self):
-        self.test_token_auth(callback=self.gl.auth)
-
-    def test_credentials_auth(self, callback=None):
-        if callback is None:
-            callback = self.gl._credentials_auth
-        token = "credauthtoken"
-        id_ = 1
-        expected = {"PRIVATE-TOKEN": token}
-
-        @urlmatch(
-            scheme="http", netloc="localhost", path="/api/v4/session", method="post"
-        )
-        def resp_cont(url, request):
-            headers = {"content-type": "application/json"}
-            content = '{{"id": {0:d}, "private_token": "{1:s}"}}'.format(
-                id_, token
-            ).encode("utf-8")
-            return response(201, content, headers, None, 5, request)
-
-        with HTTMock(resp_cont):
-            callback()
-        self.assertEqual(self.gl.private_token, token)
-        self.assertDictEqual(expected, self.gl.headers)
-        self.assertEqual(self.gl.user.id, id_)
-
     def test_token_auth(self, callback=None):
-        if callback is None:
-            callback = self.gl._token_auth
         name = "username"
         id_ = 1
 
@@ -521,10 +470,10 @@ class TestGitlab(unittest.TestCase):
             return response(200, content, headers, None, 5, request)
 
         with HTTMock(resp_cont):
-            callback()
+            self.gl.auth()
         self.assertEqual(self.gl.user.username, name)
         self.assertEqual(self.gl.user.id, id_)
-        self.assertEqual(type(self.gl.user), CurrentUser)
+        self.assertIsInstance(self.gl.user, CurrentUser)
 
     def test_hooks(self):
         @urlmatch(
@@ -537,7 +486,7 @@ class TestGitlab(unittest.TestCase):
 
         with HTTMock(resp_get_hook):
             data = self.gl.hooks.get(1)
-            self.assertEqual(type(data), Hook)
+            self.assertIsInstance(data, Hook)
             self.assertEqual(data.url, "testurl")
             self.assertEqual(data.id, 1)
 
@@ -552,7 +501,7 @@ class TestGitlab(unittest.TestCase):
 
         with HTTMock(resp_get_project):
             data = self.gl.projects.get(1)
-            self.assertEqual(type(data), Project)
+            self.assertIsInstance(data, Project)
             self.assertEqual(data.name, "name")
             self.assertEqual(data.id, 1)
 
@@ -598,7 +547,7 @@ class TestGitlab(unittest.TestCase):
 
         with HTTMock(resp_get_group):
             data = self.gl.groups.get(1)
-            self.assertEqual(type(data), Group)
+            self.assertIsInstance(data, Group)
             self.assertEqual(data.name, "name")
             self.assertEqual(data.path, "path")
             self.assertEqual(data.id, 1)
@@ -618,24 +567,208 @@ class TestGitlab(unittest.TestCase):
             self.assertEqual(data[1].id, 2)
             self.assertEqual(data[1].name, "other_name")
 
-    def test_users(self):
-        @urlmatch(
-            scheme="http", netloc="localhost", path="/api/v4/users/1", method="get"
+    @urlmatch(scheme="http", netloc="localhost", path="/api/v4/users/1", method="get")
+    def resp_get_user(self, url, request):
+        headers = {"content-type": "application/json"}
+        content = (
+            '{"name": "name", "id": 1, "password": "password", '
+            '"username": "username", "email": "email"}'
         )
-        def resp_get_user(url, request):
+        content = content.encode("utf-8")
+        return response(200, content, headers, None, 5, request)
+
+    def test_users(self):
+        with HTTMock(self.resp_get_user):
+            user = self.gl.users.get(1)
+            self.assertIsInstance(user, User)
+            self.assertEqual(user.name, "name")
+            self.assertEqual(user.id, 1)
+
+    def test_user_status(self):
+        @urlmatch(
+            scheme="http",
+            netloc="localhost",
+            path="/api/v4/users/1/status",
+            method="get",
+        )
+        def resp_get_user_status(url, request):
             headers = {"content-type": "application/json"}
-            content = (
-                '{"name": "name", "id": 1, "password": "password", '
-                '"username": "username", "email": "email"}'
-            )
+            content = '{"message": "test", "message_html": "<h1>Message</h1>", "emoji": "thumbsup"}'
             content = content.encode("utf-8")
             return response(200, content, headers, None, 5, request)
 
-        with HTTMock(resp_get_user):
+        with HTTMock(self.resp_get_user):
             user = self.gl.users.get(1)
-            self.assertEqual(type(user), User)
-            self.assertEqual(user.name, "name")
-            self.assertEqual(user.id, 1)
+        with HTTMock(resp_get_user_status):
+            status = user.status.get()
+            self.assertIsInstance(status, UserStatus)
+            self.assertEqual(status.message, "test")
+            self.assertEqual(status.emoji, "thumbsup")
+
+    def test_todo(self):
+        with open(os.path.dirname(__file__) + "/data/todo.json", "r") as json_file:
+            todo_content = json_file.read()
+            json_content = json.loads(todo_content)
+            encoded_content = todo_content.encode("utf-8")
+
+        @urlmatch(scheme="http", netloc="localhost", path="/api/v4/todos", method="get")
+        def resp_get_todo(url, request):
+            headers = {"content-type": "application/json"}
+            return response(200, encoded_content, headers, None, 5, request)
+
+        @urlmatch(
+            scheme="http",
+            netloc="localhost",
+            path="/api/v4/todos/102/mark_as_done",
+            method="post",
+        )
+        def resp_mark_as_done(url, request):
+            headers = {"content-type": "application/json"}
+            single_todo = json.dumps(json_content[0])
+            content = single_todo.encode("utf-8")
+            return response(200, content, headers, None, 5, request)
+
+        with HTTMock(resp_get_todo):
+            todo = self.gl.todos.list()[0]
+            self.assertIsInstance(todo, Todo)
+            self.assertEqual(todo.id, 102)
+            self.assertEqual(todo.target_type, "MergeRequest")
+            self.assertEqual(todo.target["assignee"]["username"], "root")
+            with HTTMock(resp_mark_as_done):
+                todo.mark_as_done()
+
+    def test_todo_mark_all_as_done(self):
+        @urlmatch(
+            scheme="http",
+            netloc="localhost",
+            path="/api/v4/todos/mark_as_done",
+            method="post",
+        )
+        def resp_mark_all_as_done(url, request):
+            headers = {"content-type": "application/json"}
+            return response(204, {}, headers, None, 5, request)
+
+        with HTTMock(resp_mark_all_as_done):
+            self.gl.todos.mark_all_as_done()
+
+    def test_deployment(self):
+        content = '{"id": 42, "status": "success", "ref": "master"}'
+        json_content = json.loads(content)
+
+        @urlmatch(
+            scheme="http",
+            netloc="localhost",
+            path="/api/v4/projects/1/deployments",
+            method="post",
+        )
+        def resp_deployment_create(url, request):
+            headers = {"content-type": "application/json"}
+            return response(200, json_content, headers, None, 5, request)
+
+        @urlmatch(
+            scheme="http",
+            netloc="localhost",
+            path="/api/v4/projects/1/deployments/42",
+            method="put",
+        )
+        def resp_deployment_update(url, request):
+            headers = {"content-type": "application/json"}
+            return response(200, json_content, headers, None, 5, request)
+
+        with HTTMock(resp_deployment_create):
+            deployment = self.gl.projects.get(1, lazy=True).deployments.create(
+                {
+                    "environment": "Test",
+                    "sha": "1agf4gs",
+                    "ref": "master",
+                    "tag": False,
+                    "status": "created",
+                }
+            )
+            self.assertEqual(deployment.id, 42)
+            self.assertEqual(deployment.status, "success")
+            self.assertEqual(deployment.ref, "master")
+
+        with HTTMock(resp_deployment_update):
+            json_content["status"] = "failed"
+            deployment.status = "failed"
+            deployment.save()
+            self.assertEqual(deployment.status, "failed")
+
+    def test_user_activate_deactivate(self):
+        @urlmatch(
+            scheme="http",
+            netloc="localhost",
+            path="/api/v4/users/1/activate",
+            method="post",
+        )
+        def resp_activate(url, request):
+            headers = {"content-type": "application/json"}
+            return response(201, {}, headers, None, 5, request)
+
+        @urlmatch(
+            scheme="http",
+            netloc="localhost",
+            path="/api/v4/users/1/deactivate",
+            method="post",
+        )
+        def resp_deactivate(url, request):
+            headers = {"content-type": "application/json"}
+            return response(201, {}, headers, None, 5, request)
+
+        with HTTMock(resp_activate), HTTMock(resp_deactivate):
+            self.gl.users.get(1, lazy=True).activate()
+            self.gl.users.get(1, lazy=True).deactivate()
+
+    def test_update_submodule(self):
+        @urlmatch(
+            scheme="http", netloc="localhost", path="/api/v4/projects/1$", method="get"
+        )
+        def resp_get_project(url, request):
+            headers = {"content-type": "application/json"}
+            content = '{"name": "name", "id": 1}'.encode("utf-8")
+            return response(200, content, headers, None, 5, request)
+
+        @urlmatch(
+            scheme="http",
+            netloc="localhost",
+            path="/api/v4/projects/1/repository/submodules/foo%2Fbar",
+            method="put",
+        )
+        def resp_update_submodule(url, request):
+            headers = {"content-type": "application/json"}
+            content = """{
+            "id": "ed899a2f4b50b4370feeea94676502b42383c746",
+            "short_id": "ed899a2f4b5",
+            "title": "Message",
+            "author_name": "Author",
+            "author_email": "author@example.com",
+            "committer_name": "Author",
+            "committer_email": "author@example.com",
+            "created_at": "2018-09-20T09:26:24.000-07:00",
+            "message": "Message",
+            "parent_ids": [ "ae1d9fb46aa2b07ee9836d49862ec4e2c46fbbba" ],
+            "committed_date": "2018-09-20T09:26:24.000-07:00",
+            "authored_date": "2018-09-20T09:26:24.000-07:00",
+            "status": null}"""
+            content = content.encode("utf-8")
+            return response(200, content, headers, None, 5, request)
+
+        with HTTMock(resp_get_project):
+            project = self.gl.projects.get(1)
+            self.assertIsInstance(project, Project)
+            self.assertEqual(project.name, "name")
+            self.assertEqual(project.id, 1)
+        with HTTMock(resp_update_submodule):
+            ret = project.update_submodule(
+                submodule="foo/bar",
+                branch="master",
+                commit_sha="4c3674f66071e30b3311dac9b9ccc90502a72664",
+                commit_message="Message",
+            )
+            self.assertIsInstance(ret, dict)
+            self.assertEqual(ret["message"], "Message")
+            self.assertEqual(ret["id"], "ed899a2f4b50b4370feeea94676502b42383c746")
 
     def _default_config(self):
         fd, temp_path = tempfile.mkstemp()
@@ -654,5 +787,5 @@ class TestGitlab(unittest.TestCase):
 
         config_path = self._default_config()
         gl = MyGitlab.from_config("one", [config_path])
-        self.assertEqual(type(gl).__name__, "MyGitlab")
+        self.assertIsInstance(gl, MyGitlab)
         os.unlink(config_path)
